import json
from abc import ABC, abstractmethod
from typing import List

import psutil
from ambient_backend_api_client.models import InterfaceTypeEnum, NetworkInterface
from result import Ok

from ambient_edge_server.utils import logger


class InterfaceService(ABC):
    @abstractmethod
    async def get_network_interfaces(self) -> List[NetworkInterface]:
        """Get network interfaces on the device."""


class LinuxInterfaceService(InterfaceService):
    def __init__(self) -> None:
        self._running = False
        self._error = Ok("initialized")

    async def get_network_interfaces(self) -> List[NetworkInterface]:
        logger.info("Starting to fetch network interface details.")
        interfaces = psutil.net_if_addrs()
        logger.debug("Interfaces: %s", json.dumps(interfaces, indent=4))
        network_interfaces = []

        for interface_name, addresses in interfaces.items():
            logger.debug("Processing %s interface ...", interface_name)
            logger.debug("Addresses: %s", addresses)

            for addr in addresses:
                details = {
                    "name": f"{interface_name}-{addr.address}",
                    # Default to OTHER, adjust logic as needed
                    "type": InterfaceTypeEnum.UNKNOWN,
                }
                logger.debug("Address: %s", addr)
                if addr.family == 2:
                    details["ipv4_address"] = addr.address
                    details["netmask"] = addr.netmask
                elif addr.family == 10:
                    details["ipv6_address"] = addr.address
                details["broadcast"] = addr.broadcast

                # Interface type heuristic
                wifi_heuristics = ["wi-fi", "wifi", "wlan"]
                ethernet_heuristics = [
                    "eth",
                    "enp",
                    "ens",
                    "eno",
                    "enx",
                    "docker",
                    "lo",
                    "br",
                ]

                if any(
                    heuristic in interface_name.lower() for heuristic in wifi_heuristics
                ):
                    details["type"] = InterfaceTypeEnum.WIFI
                elif any(
                    heuristic in interface_name.lower()
                    for heuristic in ethernet_heuristics
                ):
                    details["type"] = InterfaceTypeEnum.ETHERNET

                network_interface = NetworkInterface(**details)
                logger.debug(
                    "Network Interface: %s", network_interface.model_dump_json(indent=4)
                )
                network_interfaces.append(network_interface)
            logger.info(f"Added {interface_name} to network interfaces.")

        logger.info("Completed fetching all network interfaces.")
        return network_interfaces



from ambient_backend_api_client import Configuration
from ambient_event_bus_client import Client, ClientOptions
from docker import DockerClient

from ambient_edge_server.config import settings
from ambient_edge_server.repos.node_repo import FileNodeRepo
from ambient_edge_server.repos.system_daemon_repo import LinuxSystemDaemonRepo
from ambient_edge_server.repos.token_repo import EncryptedTokenRepository
from ambient_edge_server.services import (
    authorization_service,
    event_service,
    handler_service,
    health_service,
    interface_service,
    port_service,
    system_daemon_service,
)
from ambient_edge_server.utils import logger


class ServiceManager:
    def __init__(self):
        if settings.platform == "linux":
            self._system_daemon_repo = LinuxSystemDaemonRepo()
            self._system_daemon_service = system_daemon_service.LinuxDaemonService(
                self._system_daemon_repo
            )
            self._interface_service = interface_service.LinuxInterfaceService()
        else:
            logger.error("Platform not supported")
            raise NotImplementedError("Platform not supported")
        self._port_service = port_service.PortService()
        self._token_repo = EncryptedTokenRepository()
        self._node_repo = FileNodeRepo()
        self._authorization_service = authorization_service.AuthorizationService(
            token_repo=self._token_repo, node_repo=self._node_repo
        )

        self._event_client = Client(
            ClientOptions(
                event_api_url=settings.event_bus_api,
                connection_service_url=settings.connection_service_url,
                api_token="dummy_token",
                log_level="DEBUG",
            )
        )
        self._event_service = event_service.AmbientBusEventService(
            client=self._event_client, node_repo=self._node_repo
        )

        self._docker_client = DockerClient()
        self._handler_service = handler_service.HandlerService(
            self._event_service, docker_client=self._docker_client
        )

        # Keep Health Service at the bottom since
        # it should be able to access all other services
        if settings.platform == "linux":
            self._health_service = health_service.LinuxHealthService(
                node_repo=self._node_repo,
                event_service=self._event_service,
                auth_service=self._authorization_service,
                interface_service=self._interface_service,
                interval_minutes=5,
            )

    async def init(self):
        await self._port_service.init()

        token = await self._authorization_service.get_token()
        if not token:
            logger.error("Failed to get token")
            return
        logger.debug("got token [ %d chars ]", len(token))
        self._event_client.token = token
        await self._event_client.init_client()
        await self._event_service.start()
        await self._handler_service.start()

        # Keep Health Service at the bottom since
        # it should be able to access all other services
        api_config = Configuration(
            host=settings.backend_api_url,
            access_token=token,
        )
        await self._health_service.start(api_config=api_config)

    def get_port_service(self):
        return self._port_service

    def get_authorization_service(self):
        return self._authorization_service

    def get_event_service(self):
        return self._event_service

    def get_system_daemon_service(self):
        return self._system_daemon_service

    def get_interface_service(self):
        return self._interface_service


svc_manager = ServiceManager()



from fastapi import FastAPI

from ambient_edge_server.routers import auth, daemon, ping, ports
from ambient_edge_server.services.service_manager import svc_manager
from ambient_edge_server.utils import logger

app = FastAPI()


async def startup():
    logger.info("Starting up ...")
    try:
        await svc_manager.init()
        logger.info("Services initialized.")
    except Exception as e:
        logger.warning("Failed to initialize services: %s", e)


app.add_event_handler("startup", startup)

routers = [ping, ports, auth, daemon]

for router in routers:
    app.include_router(router.router)



import platform

from pydantic_settings import BaseSettings


class Settings(BaseSettings):
    ambient_log_level: str = "DEBUG"
    ambient_config: str = "~/.ambientctl/config.json"

    backend_api_url: str = "https://api.ambientlabs.io"
    event_bus_api: str = "https://events.ambientlabs.io"
    connection_service_url: str = "wss://sockets.ambientlabs.io"

    service_template_location: str = (
        "./ambient_edge_server/templates/ambient_edge_server.service.jinja2"
    )
    platform: str = platform.system().lower()


settings = Settings()



import os
import sys

import uvicorn


def run():
    print(f"sys.path: {sys.path}    os.getcwd(): {os.getcwd()}")
    try:
        from ambient_edge_server.app import app
        from ambient_edge_server.utils import logger

        logger.debug("running server ...")
        logger.debug("running app: %s ...", app)
        uvicorn.run(app, host="0.0.0.0", port=7417)
    except Exception as e:
        print("error running server: %s", e)


if __name__ == "__main__":
    run()



import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Union

from ambient_backend_api_client import Node

from ambient_edge_server.utils import logger


class NodeRepo(ABC):
    @abstractmethod
    def save_node_data(self, node: Node) -> None:
        """Save the node data to the repository.

        Args:
            node (Node): Node data
        """

    @abstractmethod
    def get_node_data(self, strict: bool = False) -> Union[Node, None]:
        """Get the node data from the repository.

        Returns:
            Node: Node data or None if not found
            strict (bool): If True, raise an exception if the node data is not found
        """

    @abstractmethod
    def get_node_id(self) -> int:
        """Get the node ID.

        Returns:
            int: Node ID
        """


class FileNodeRepo(NodeRepo):
    def __init__(self):
        self.home_path = Path.home() / ".ambient"
        self.node_data_path = self.home_path / ".node_data.json"

        self._init_files()

    def _init_files(self):
        if not self.node_data_path.exists():
            self._write_to_file(self.node_data_path, b"")

    def _write_to_file(self, file_path: Path, data: bytes):
        with open(file_path, "wb") as f:
            f.write(data)

    def save_node_data(self, node: Node) -> None:
        self._write_to_file(
            self.node_data_path, node.model_dump_json(indent=4).encode()
        )

    def get_node_id(self) -> Union[int, None]:
        logger.debug("Reading node ID from file: %s", self.node_data_path)
        node_data = self.get_node_data()
        if node_data:
            logger.debug("Node ID: %s", node_data.id)
            return node_data.id
        return os.getenv("AMBIENT_NODE_ID", None)

    def get_node_data(self, strict: bool = False) -> Union[Node, None]:
        logger.debug("Reading node data from file: %s", self.node_data_path)
        with open(self.node_data_path, "rb") as f:
            contents = f.read().decode()
            logger.debug("Node data: %s", contents)
            if strict:
                return Node.model_validate_json(contents)
            if not contents or contents == "":
                logger.info("No node data found")
                return None
            return Node.model_validate_json(contents)



import asyncio
from abc import ABC, abstractmethod
from typing import Callable, Set, Union

from ambient_backend_api_client import EventLabel
from ambient_event_bus_client import Client, Message, MessageCreate
from result import Err, Result

from ambient_edge_server.repos.node_repo import NodeRepo
from ambient_edge_server.utils import logger


class EventService(ABC):
    @abstractmethod
    async def start(self) -> None:
        """Start the event service."""

    @abstractmethod
    async def stop(self) -> None:
        """Stop the event service."""

    @abstractmethod
    async def add_event_handler(
        self, event_label: EventLabel, handler: Callable
    ) -> None:
        """Add an event handler to the event service.

        Args:
            event_label (EventLabel): Event label
            handler (Callable): async function to handle the event
        """

    @abstractmethod
    async def send_event(self, topic: str, msg: str) -> None:
        """Send an event to the event service.

        Args:
            event (Event): Event to send
        """

    @property
    @abstractmethod
    async def is_running(self) -> bool:
        """Check if the event service is running.

        Returns:
            bool: True if the event service is running, False otherwise
        """

    @property
    @abstractmethod
    def error(self) -> Union[Result, None]:
        """Get the current error state of the event service.

        Returns:
            Result: The current error state
        """


class AmbientBusEventService(EventService):
    def __init__(self, client: Client, node_repo: NodeRepo) -> None:
        self.client = client
        self.node_repo = node_repo
        self.handlers = {}
        self.tasks: Set[asyncio.Task] = set()
        self._error = None
        self._is_connected = False

    async def start(self) -> None:
        logger.info("Starting event service ...")
        await self.client.init_client()
        node = self.node_repo.get_node_data()
        if node:
            await self.client.add_subscription(f"node-{node.id}/*")
        logger.debug("event client initialized.")

        logger.info("Starting subscription loop ...")
        sub_loop_task = asyncio.create_task(self._subscription_loop())
        sub_loop_task.add_done_callback(self._done_callback)
        sub_loop_task.set_name("subscription_loop_task")
        self.tasks.add(sub_loop_task)
        logger.info("subscription loop started.")

    async def stop(self) -> None:
        logger.info("Stopping event service ...")
        for task in self.tasks:
            logger.debug("cancelling task: %s", task.get_name())
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass

    async def add_event_handler(
        self, event_label: EventLabel, handler: Callable
    ) -> None:
        logger.info("Adding event handler for event: %s", event_label.value)

        logger.debug(
            "handler name: %s [ from %s ]", handler.__name__, handler.__module__
        )
        self.handlers[event_label.value] = handler

    async def send_event(self, topic: str, msg: str) -> None:
        message_data = MessageCreate(topic=topic, message=msg)
        logger.debug("Publishing message:\n%s", message_data.model_dump_json(indent=4))
        await self.client.publish(msg)

    @property
    async def is_running(self) -> bool:
        return all(task.done() for task in self.tasks) and self._is_connected

    @property
    def error(self) -> Union[Err, None]:
        return self._error

    async def _get_handler(self, topic: str) -> Callable:
        logger.debug("Getting handler for topic: %s", topic)
        handler_items = self.handlers.items()
        logger.debug("handlers: %s", handler_items)
        for event_label, handler in handler_items:
            if event_label in topic:
                logger.debug("Handler found for topic: %s", topic)
                return handler
        logger.info("No handler found for topic: %s", topic)
        return None

    async def _handle_event(self, msg: Message) -> None:
        handler = await self._get_handler(msg.topic)
        if handler:
            try:
                await handler(msg)
            except Exception as e:
                self._error = Err(e)
        else:
            self._error = Err(f"No handler for event: {msg.topic}")

    async def _subscription_loop(self) -> None:
        logger.info("Starting subscription loop ...")
        while True:
            self._error = None
            self._is_connected = True
            try:
                async for message in self.client.subscribe():
                    logger.debug(
                        "Received message:\n%s", message.model_dump_json(indent=4)
                    )
                    await self._handle_event(message)
            except Exception as e:
                self._error = Err(e)

    async def _done_callback(self, task: asyncio.Task) -> None:
        err_msg = f"task {task.get_name()} ended. \
Result: {task.result() if task.done() else task.exception()}"
        logger.warning(err_msg)
        self._error = Err(err_msg)

        logger.debug("removing task from tasks set ...")
        self.tasks.remove(task)

        logger.info("restarting subscription loop ...")
        sub_loop_task = asyncio.create_task(self._subscription_loop())
        sub_loop_task.add_done_callback(self._done_callback)
        sub_loop_task.set_name("subscription_loop_task")
        self.tasks.add(sub_loop_task)
        logger.info("subscription loop restarted.")
